# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""Tools to visualize neuron morphological objects.

Examples:
    >>> from neurom import viewer
    >>> nrn = ... # load a neuron
    >>> viewer.draw(nrn)                    # 2d plot
    >>> viewer.draw(nrn, mode='3d')         # 3d plot
    >>> viewer.draw(nrn.neurites[0])        # 2d plot of neurite tree
    >>> viewer.draw(nrn, mode='dendrogram') # dendrogram plot
"""

from .view import (plot_neuron, plot_neuron3d,
                   plot_tree, plot_tree3d,
                   plot_soma, plot_soma3d,
                   plot_dendrogram)
from .view import common
from .core import Tree, Neurite, Soma, Neuron


MODES = ('2d', '3d', 'dendrogram')

_VIEWERS = {
    'neuron_3d': plot_neuron3d,
    'neuron_2d': plot_neuron,
    'neuron_dendrogram': plot_dendrogram,
    'tree_3d': plot_tree3d,
    'tree_2d': plot_tree,
    'tree_dendrogram': plot_dendrogram,
    'soma_3d': plot_soma3d,
    'soma_2d': plot_soma
}


class ViewerError(Exception):
    """Base class for viewer exceptions."""


class InvalidDrawModeError(ViewerError):
    """Exception class to indicate invalid draw mode."""


class NotDrawableError(Exception):
    """Exception class for things that aren't drawable."""


def draw(obj, mode='2d', **kwargs):
    """Draw a morphology object.

    Arguments:
        obj: morphology object to be drawn (neuron, tree, soma).
        mode (Optional[str]): drawing mode ('2d', '3d', 'dendrogram'). Defaults to '2d'.
        **kwargs: keyword arguments for underlying neurom.view.view functions.

    Raises:
        InvalidDrawModeError if mode is not valid
        NotDrawableError if obj is not drawable
        NotDrawableError if obj type and mode combination is not drawable

    Examples:
        >>> nrn = ... # load a neuron
        >>> fig, _ = viewer.draw(nrn)             # 2d plot
        >>> fig.show()
        >>> fig3d, _ = viewer.draw(nrn, mode='3d') # 3d plot
        >>> fig3d.show()
        >>> fig, _ = viewer.draw(nrn.neurites[0]) # 2d plot of neurite tree
        >>> dend, _ = viewer.draw(nrn, mode='dendrogram')
    """
    if mode not in MODES:
        raise InvalidDrawModeError('Invalid drawing mode %s' % mode)

    if 'realistic_diameters' in kwargs and mode == '3d':
        if kwargs['realistic_diameters']:
            raise NotImplementedError('Option realistic_diameter not implemented for 3D plots')
        del kwargs['realistic_diameters']

    fig, ax = (common.get_figure() if mode in ('2d', 'dendrogram')
               else common.get_figure(params={'projection': '3d'}))

    if isinstance(obj, Neuron):
        tag = 'neuron'
    elif isinstance(obj, (Tree, Neurite)):
        tag = 'tree'
    elif isinstance(obj, Soma):
        tag = 'soma'
    else:
        raise NotDrawableError('draw not implemented for %s' % obj.__class__)

    viewer = '%s_%s' % (tag, mode)
    try:
        plotter = _VIEWERS[viewer]
    except KeyError:
        raise NotDrawableError('No drawer for class %s, mode=%s' % (obj.__class__, mode))

    output_path = kwargs.pop('output_path', None)
    plotter(ax, obj, **kwargs)

    if mode != 'dendrogram':
        common.plot_style(fig=fig, ax=ax, **kwargs)

    if output_path:
        common.save_plot(fig=fig, output_path=output_path, **kwargs)

    return fig, ax



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path

from nose import tools as nt
from neurom.core.population import Population
from neurom import load_neuron

DATA_PATH = Path(__file__).parent.parent.parent.parent / 'test_data'

NRN1 = load_neuron(Path(DATA_PATH, 'swc/Neuron.swc'))
NRN2 = load_neuron(Path(DATA_PATH, 'swc/Single_basal.swc'))
NRN3 = load_neuron(Path(DATA_PATH, 'swc/Neuron_small_radius.swc'))

NEURONS = [NRN1, NRN2, NRN3]
TOT_NEURITES = sum(len(N.neurites) for N in NEURONS)
POP = Population(NEURONS, name='foo')


def test_population():
    nt.assert_equal(len(POP.neurons), 3)
    nt.ok_(POP.neurons[0].name, 'Neuron')
    nt.ok_(POP.neurons[1].name, 'Single_basal')
    nt.ok_(POP.neurons[2].name, 'Neuron_small_radius')

    nt.assert_equal(len(POP.somata), 3)

    nt.assert_equal(len(POP.neurites), TOT_NEURITES)

    nt.assert_equal(POP.name, 'foo')


def test_neurons():
    for i, n in enumerate(NEURONS):
        nt.assert_true(n is POP.neurons[i])


def test_iterate_neurons():
    for a, b in zip(NEURONS, POP):
        nt.assert_true(a is b)


def test_len():
    nt.assert_equal(len(POP), len(NEURONS))


def test_getitem():
    for i in range(len(NEURONS)):
        nt.assert_true(POP[i] is NEURONS[i])


def test_str():
    nt.ok_('Population' in str(POP))



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test neurom._bifurcationfunc functionality."""

from pathlib import Path
import warnings

import numpy as np
from nose import tools as nt
from nose.tools import assert_equal, assert_raises
from numpy.testing import assert_raises

import neurom as nm
from neurom import load_neuron
from neurom.exceptions import NeuroMError

# NOTE: The 'bf' alias is used in the fst/tests modules
# Do NOT change it.
# TODO: If other neurom.features are imported,
# the should use the aliasing used in fst/tests module files
from neurom.features import bifurcationfunc as bf

DATA_PATH = Path(__file__).parent / '../../../test_data/'
SWC_PATH = Path(DATA_PATH, 'swc')
SIMPLE = nm.load_neuron(Path(SWC_PATH, 'simple.swc'))
with warnings.catch_warnings(record=True):
    SIMPLE2 = load_neuron(Path(DATA_PATH, 'neurolucida', 'not_too_complex.asc'))
    MULTIFURCATION = load_neuron(Path(DATA_PATH, 'neurolucida', 'multifurcation.asc'))


def test_local_bifurcation_angle():
    nt.ok_(bf.local_bifurcation_angle(SIMPLE.sections[0]) == np.pi)
    nt.ok_(bf.local_bifurcation_angle(SIMPLE.sections[3]) == np.pi)
    leaf = SIMPLE.section(2)
    assert_raises(NeuroMError, bf.local_bifurcation_angle, leaf)


def test_remote_bifurcation_angle():
    nt.ok_(bf.remote_bifurcation_angle(SIMPLE.sections[0]) == np.pi)
    nt.ok_(bf.remote_bifurcation_angle(SIMPLE.sections[3]) == np.pi)
    leaf = SIMPLE.section(2)
    assert_raises(NeuroMError, bf.local_bifurcation_angle, leaf)


def test_bifurcation_partition():
    root = SIMPLE2.neurites[0].root_node
    assert_equal(bf.bifurcation_partition(root), 3.0)
    assert_equal(bf.bifurcation_partition(root.children[0]), 1.0)

    leaf = root.children[0].children[0]
    assert_raises(NeuroMError, bf.bifurcation_partition, leaf)

    multifurcation_section = MULTIFURCATION.neurites[0].root_node.children[0]
    assert_raises(NeuroMError, bf.bifurcation_partition, multifurcation_section)


def test_partition_asymmetry():
    root = SIMPLE2.neurites[0].root_node
    assert_equal(bf.partition_asymmetry(root), 0.5)
    assert_equal(bf.partition_asymmetry(root.children[0]), 0.0)

    leaf = root.children[0].children[0]
    assert_raises(NeuroMError, bf.partition_asymmetry, leaf)

    multifurcation_section = MULTIFURCATION.neurites[0].root_node.children[0]
    assert_raises(NeuroMError, bf.partition_asymmetry, multifurcation_section)


def test_sibling_ratio():
    root = SIMPLE2.neurites[0].root_node
    assert_equal(bf.sibling_ratio(root), 1.0)
    assert_equal(bf.sibling_ratio(root.children[0]), 1.0)

    assert_equal(bf.sibling_ratio(root, method='mean'), 1.0)
    assert_equal(bf.sibling_ratio(root.children[0], method='mean'), 1.0)

    leaf = root.children[0].children[0]
    assert_raises(NeuroMError, bf.sibling_ratio, leaf)
    assert_raises(NeuroMError, bf.sibling_ratio, leaf, method='mean')

    multifurcation_section = MULTIFURCATION.neurites[0].root_node.children[0]
    assert_raises(NeuroMError, bf.sibling_ratio, multifurcation_section)

    assert_raises(ValueError, bf.sibling_ratio, root, method='unvalid-method')


def test_diameter_power_relation():
    root = SIMPLE2.neurites[0].root_node
    assert_equal(bf.diameter_power_relation(root), 2.0)
    assert_equal(bf.diameter_power_relation(root.children[0]), 2.0)

    assert_equal(bf.diameter_power_relation(root, method='mean'), 2.0)
    assert_equal(bf.diameter_power_relation(root.children[0], method='mean'), 2.0)

    leaf = root.children[0].children[0]
    assert_raises(NeuroMError, bf.diameter_power_relation, leaf)
    assert_raises(NeuroMError, bf.diameter_power_relation, leaf, method='mean')

    multifurcation_section = MULTIFURCATION.neurites[0].root_node.children[0]
    assert_raises(NeuroMError, bf.diameter_power_relation, multifurcation_section)

    assert_raises(ValueError, bf.diameter_power_relation, root, method='unvalid-method')
